// QConnection.cpp: implementation of the QConnection class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Query32.h"
#include "QConnection.h"
#include "QDN.h"

#include "ldap-int.h"
#define LDAP_SETCLR_BITOPT( ld, bit, optdata ) \
	if ( optdata != NULL ) {		\
		(ld)->ld_options |= bit;	\
	} else {				\
		(ld)->ld_options &= ~bit;	\
	}

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

QConnection::QConnection
	( long conn_id_arg,
		CString &host_arg, int port_arg,
		CString &user_arg, CString &password_arg )
{
  ref_count = 0;
  conn_id = conn_id_arg;

  ldapctx = oldldapctx = NULL;
  bound = binding = FALSE;

  host = host_arg;
	port = port_arg;

  user = user_arg;
  password = password_arg;
}

QConnection::~QConnection()
{
  if ( bound )
    Unbind();
}

//////////////////////////////////////////////////////////////////////
// QConnection::Compare
// Compare this connection to the given LDAP connection specifiers
// (host, port, username and password)
BOOL QConnection::Compare
  ( const CString &host, int port, const CString &user, const CString &password )
{
  WIN32_GUFF();

  if ( this->port == port && this->host.CompareNoCase( host ) == 0 )
    return TRUE;

  return FALSE;
}

//////////////////////////////////////////////////////////////////////
// QConnection::RebindAsync
// Rebind if not already connected. Cleans up previous LDAP context.
BOOL QConnection::RebindAsync( QError *&err, int nLdapVersion /* = LDAP_VERSION3 */ )
{
  WIN32_GUFF();

	if ( Bound() == TRUE )
		return TRUE;

	int lerr;
  if ( binding == TRUE )
    //
    // A bind attempt is already in progress!
    {
      err = new QError( QERR_API_ERROR );
      return FALSE;
    }
  else if ( host.IsEmpty() )
    //
    // Can't bind without a directory server address!
    {
      err = new QError( QERR_INVALID_ARGUMENT );
      return FALSE;
    }
  else
    {
      binding = TRUE;

      //
      // Get LDAP connection structure
      if ( ldapctx != ( LDAP * ) NULL )
	      {
	        oldldapctx = ldapctx;
	        ldapctx = ( LDAP * ) NULL;
	      }

			char *tmp_host = strdup( host );
      ldapctx = ldap_init( host, port );
			free( tmp_host );

      if ( ldapctx == ( LDAP * ) NULL )
	      {
					QLOG( " *** QConnection::RebindAsync()" );
					QLOGRAW( "Failed to bind to LDAP server (at ldap_init):" );
					QLOGRAW( "Host: %s\nPort: %d", host, port );

	        lerr = LDAP_OPERATIONS_ERROR;
	        err = new QError( QERR_INVALID_ARGUMENT );

	        if (oldldapctx != ( LDAP * ) NULL)
	          {
	            ldapctx = oldldapctx;
	            oldldapctx = ( LDAP * ) NULL;
	          }

					return FALSE;
      	}
      else
	      {
          // Set LDAP protocol version
          ldap_set_option( ldapctx, LDAP_OPT_PROTOCOL_VERSION, (void *) &nLdapVersion );

          // Automatic referral chasing off
					LDAP_SETCLR_BITOPT( ldapctx, LDAP_BITOPT_REFERRALS, LDAP_OPT_OFF );

	        // Try binding to directory server
					char *tmp_user = strdup( user );
					char *tmp_passwd = strdup( password );

					bind_msgid = ldap_simple_bind( ldapctx, user, password );

					free( tmp_user );
					free( tmp_passwd );

					if ( bind_msgid <= 0 )
						{
							QLOG( " *** QConnection::RebindAsync()" );
							QLOGRAW( "Failed to bind to LDAP server (at ldap_simple_bind):" );
							QLOGRAW( "Host: %s\nPort: %d", host, port );

							err = new QError( QERR_BIND_FAILED );
							QDN tmpDN;
							err->SetOpInfo( QERR_OP_BIND, tmpDN, host, port );
							return FALSE;
						}
	      }
    }

	UpdateLastActive();
  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// QConnection::BindAsync
// Perform an asynchronous bind. Assumes this connection hasn't
// previously bound.
BOOL QConnection::BindAsync( QError *&err, int nLdapVersion /* = LDAP_VERSION3 */  )
{
  WIN32_GUFF();

  int lerr;
  if ( binding == TRUE )
    //
    // A bind attempt is already in progress!
    {
      err = new QError( QERR_API_ERROR );
      return FALSE;
    }
  else if ( host.IsEmpty() || host == "" )
    //
    // Can't bind without a directory server address!
    {
      err = new QError( QERR_INVALID_ARGUMENT );
      return FALSE;
    }
  else
    {
      binding = TRUE;

      //
      // Get LDAP connection structure
      if ( ldapctx != ( LDAP * ) NULL )
	      {
	        oldldapctx = ldapctx;
	        ldapctx = ( LDAP * ) NULL;
	      }

      ldapctx = ldap_init( host, port );
      if ( ldapctx == ( LDAP * ) NULL )
	      {
					QLOG( " *** QConnection::BindAsync()" );
					QLOGRAW( "Failed to bind to LDAP server (at ldap_init):" );
					QLOGRAW( "Host: %s\nPort: %d", host, port );

					lerr = LDAP_OPERATIONS_ERROR;
	        err = new QError( QERR_INVALID_ARGUMENT );

	        if (oldldapctx != ( LDAP * ) NULL)
	          {
	            ldapctx = oldldapctx;
	            oldldapctx = ( LDAP * ) NULL;
	          }

					return FALSE;
      	}
      else
	      {
          // Set LDAP protocol version 
          ldap_set_option( ldapctx, LDAP_OPT_PROTOCOL_VERSION, &nLdapVersion );

          // Automatic referral chasing off
					LDAP_SETCLR_BITOPT( ldapctx, LDAP_BITOPT_REFERRALS, LDAP_OPT_OFF );
					
	        // Try binding to directory server
	        bind_msgid = ldap_simple_bind( ldapctx, user, password );
          if ( bind_msgid < 0 )
						{
							QLOG( " *** QConnection::BindAsync()" );
							QLOGRAW( "Failed to bind to LDAP server (at ldap_simple_bind):" );
							QLOGRAW( "Host: %s\nPort: %d", host, port );

							err = new QError( QERR_BIND_FAILED );
							QDN tmpDN;
							err->SetOpInfo( QERR_OP_BIND, tmpDN, host, port );
							return FALSE;
						}
	      }
    }

	UpdateLastActive();
  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// QConnection::BindWait
// Wait for an asynchronous bind operation on this QConnection to
// complete.
BOOL QConnection::BindWait( QError *&err )
{
  WIN32_GUFF();

	UpdateLastActive();
	if ( Bound() == TRUE )
		// No need to wait if already bound
		return TRUE;

  if ( binding == FALSE )
    // Can't wait for bind to return if hasn't been initiated!
    {
      err = new QError( QERR_INTERNAL );
      return FALSE;
    }

  // Wait for bind confirmation
  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = 10;

  // While the operation is still running, do this:
  LDAPMessage *lm = NULL;
  int rc = 0;

  while ( rc == 0 && binding == TRUE )
    {
      // Check the status of the LDAP operation
      rc = ldap_result( ldapctx, bind_msgid, 0, &timeout, &lm );
      switch ( rc )
        {
          // If -1 was returned, an error occurred
          case -1:
             break;

          // If 0 was returned, the operation is still in progress
          case 0:
             continue;

          // If any other value is returned, assume we are done
          default:
            // Check if the "bind" operation was successful
						int rc = ldap_result2error( ldapctx, lm , 0 );
            if ( rc != LDAP_SUCCESS )
              {
								char *pLdapErr = ldap_err2string( rc );

								QLOG( "QConnection::BindWait()" );
								QLOGRAW( "Failed to bind to LDAP server:" );
								QLOGRAW( "LDAP Error: %s", pLdapErr );
								QLOGRAW( "Host: %s", host );
								QLOGRAW( "Port: %d", port );

								//
								// If we're getting a protocol error retry with
								// LDAP_VERSION2 in case it's down to a version mismatch
								if ( rc = LDAP_PROTOCOL_ERROR )
									{
										QLOG( "QConnection::BindWait" );
										QLOGRAW( "Attempting to bind using LDAP_VERSION2." );
										QLOGRAW( "Host: %s", host );
										QLOGRAW( "Port: %d", port );

										binding = FALSE;
										RebindAsync( err, LDAP_VERSION2 );
									}
								else
									{
										binding = FALSE;
										err = new QError( QERR_LOST_SERVER_CONNECTION );
										QDN tmpDN;
										err->SetOpInfo( QERR_OP_BIND, tmpDN, host, port );
										return( FALSE );
									}
              }
        }

      // Looks like it's going to take some time,
      // so set non-blocking bind-wait.
      timeout.tv_sec = 0;

      if ( binding == FALSE )
	      // Bind request has been cancelled...
	      rc = ldap_abandon( ldapctx, bind_msgid );
    }

  //
  // Check for errors...
  if ( rc == -1 )
    {
      rc = ldap_result2error( ldapctx, lm, FALSE );
      if ( rc != LDAP_SUCCESS )
	      {
					binding = FALSE;

					err = new QError();
					err->SetLdapError( rc, lm );

					QDN tmpDN;
					err->SetOpInfo( QERR_OP_BIND, tmpDN, host, port );

					(void) ldap_msgfree( lm );
	        return FALSE;
	      }
    }

  if ( lm != NULL )
    ldap_msgfree( lm );

  if ( binding == TRUE )
    bound = TRUE;

  //
  // Unbind old connection
  if ( oldldapctx != (LDAP *) NULL )
    {
      ldap_unbind( oldldapctx );
      oldldapctx = (LDAP *) NULL;
    }

	//
	// Setup socket notification
	SOCKET socket;
	ldap_get_option( ldapctx, LDAP_OPT_DESC, &socket );

	CWnd *pMainWnd = AfxGetMainWnd();
	WSAAsyncSelect( socket, pMainWnd->GetSafeHwnd(), PG_CLOSE_EVENT, FD_CLOSE );
	
	UpdateLastActive();
  binding = FALSE;
  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// QConnection::Bind XXX Unimplemented
// Perform a synchronous bind on this QConnection. 
BOOL QConnection::Bind( QError *&err )
{
  return FALSE;
}

//////////////////////////////////////////////////////////////////////
// QConnection::Unbind
// Unbind on this connection.
void QConnection::Unbind( void )
{
  WIN32_GUFF();

  ldap_unbind(ldapctx);
  ldapctx = (LDAP *) NULL;
  bound = FALSE;
}

//////////////////////////////////////////////////////////////////////
// QConnection::SelectRead
// Select on read events.
int QConnection::SelectRead( struct timeval *timeout )
{
  WIN32_GUFF();

	struct fd_set readfd;
	static int tblsize = 0;

  if ( ldapctx == NULL )
    return 0;

	if ( tblsize == 0 )
		tblsize = 64;

	SOCKET sock_id;
	ldap_get_option( ldapctx, LDAP_OPT_DESC, &sock_id );
	
	FD_ZERO( &readfd );
	FD_SET( sock_id, &readfd );

	UpdateLastActive();
	return ( select( tblsize, &readfd, NULL, NULL, timeout ) );
}

//////////////////////////////////////////////////////////////////////
// QConnection::SelectExceptions
// Select on exceptions.
int QConnection::SelectExceptions( struct timeval *timeout )
{
  WIN32_GUFF();

	struct fd_set exceptfd;
	static int tblsize = 0;

  if ( ldapctx == NULL )
    return 0;

	if ( tblsize == 0 )
		tblsize = 64;

	SOCKET sock_id;
	ldap_get_option( ldapctx, LDAP_OPT_DESC, &sock_id );
	
	FD_ZERO( &exceptfd );
	FD_SET( sock_id, &exceptfd );

	UpdateLastActive();
	return ( select( tblsize, NULL, NULL, &exceptfd, timeout ) );
}

//////////////////////////////////////////////////////////////////////
// QConnection::OnClose
// Check if connection is bound.
BOOL QConnection::Bound()
{
	return bound;
}

//////////////////////////////////////////////////////////////////////
// QConnection::Socket
// Return the socket associated with this server connection.
SOCKET QConnection::Socket()
{
	SOCKET socket;
	ldap_get_option( ldapctx, LDAP_OPT_DESC, &socket );
	return socket;
}

//////////////////////////////////////////////////////////////////////
// QConnection::OnClose
// Notifier called when a socket has been closed
void QConnection::OnClose( int nErrorCode )
{
	if ( bound == TRUE )
		{
			//
			// Connection closed unexpectedly!
			Unbind();
		}
}

//////////////////////////////////////////////////////////////////////
// QConnection::Dump
// Dump connection details to the log file
void QConnection::Dump()
{
	WIN32_GUFF();
	QLOGRAW( "Host: %s\nPort: %d", host, port );
}
